import os
import gzip
from typing import Union
import pytest
import numpy as np
import cramjam
import hashlib
from hypothesis import strategies as st, given, settings
from hypothesis.extra import numpy as st_np


VARIANTS = (
    "snappy",
    "brotli",
    "bzip2",
    "lz4",
    "gzip",
    "deflate",
    "zstd",
    "zlib",
    "xz",
)
for experimental_feat in ("blosc2", "igzip", "ideflate", "izlib"):
    if not hasattr(cramjam, experimental_feat) and hasattr(cramjam, "experimental"):
        mod = getattr(cramjam.experimental, experimental_feat, None)
        if mod:
            setattr(cramjam, experimental_feat, mod)
            VARIANTS = (*VARIANTS, experimental_feat)

# Some OS can be slow or have higher variability in their runtimes on CI
settings.register_profile("local", deadline=None, max_examples=20)
settings.register_profile("CI", deadline=None, max_examples=10)
if os.getenv("CI"):
    settings.load_profile("CI")
else:
    settings.load_profile("local")


def same_same(a, b):
    return hashlib.md5(a).hexdigest() == hashlib.md5(b).hexdigest()


def test_has_version():
    from cramjam import __version__

    assert isinstance(__version__, str)


@pytest.mark.parametrize("variant_str", VARIANTS)
@given(arr=st_np.arrays(st_np.scalar_dtypes(), shape=st.integers(0, int(1e4))))
def test_variants_different_dtypes(variant_str, arr, is_pypy):
    variant = getattr(cramjam, variant_str)
    compressed = variant.compress(arr)
    decompressed = variant.decompress(compressed)
    assert same_same(bytes(decompressed), arr.tobytes())

    # And compress n dims > 1
    if arr.shape[0] % 2 == 0:
        arr = arr.reshape((2, -1))

        if is_pypy:
            try:
                compressed = variant.compress(arr)
            except:
                pytest.skip(
                    reason="PyPy struggles w/ multidim buffer views depending on dtype ie datetime[64]"
                )
        else:
            compressed = variant.compress(arr)
        decompressed = variant.decompress(compressed)
        assert same_same(bytes(decompressed), arr.tobytes())


@pytest.mark.parametrize("is_bytearray", (True, False))
@pytest.mark.parametrize("variant_str", VARIANTS)
@given(uncompressed=st.binary(min_size=1))
def test_variants_simple(variant_str, is_bytearray, uncompressed):
    variant = getattr(cramjam, variant_str)

    if is_bytearray:
        uncompressed = bytearray(uncompressed)

    compressed = variant.compress(uncompressed)
    assert compressed.read() != uncompressed
    compressed.seek(0)
    assert isinstance(compressed, cramjam.Buffer)

    decompressed = variant.decompress(compressed, output_len=len(uncompressed))
    assert same_same(decompressed.read(), uncompressed)
    assert isinstance(decompressed, cramjam.Buffer)


@pytest.mark.parametrize("variant_str", VARIANTS)
def test_variants_raise_exception(variant_str):
    variant = getattr(cramjam, variant_str)
    with pytest.raises(cramjam.DecompressionError):
        variant.decompress(b"sknow")


@pytest.mark.parametrize(
    "input_type", (bytes, bytearray, "numpy", cramjam.Buffer, cramjam.File, memoryview)
)
@pytest.mark.parametrize(
    "output_type", (bytes, bytearray, "numpy", cramjam.Buffer, cramjam.File, memoryview)
)
@pytest.mark.parametrize("variant_str", VARIANTS)
@given(raw_data=st.binary())
def test_variants_compress_into(
    variant_str,
    input_type,
    output_type,
    raw_data,
    tmp_path_factory,
    is_pypy,
    is_free_threaded,
):
    # TODO: Fix segfault when using blosc2 compress_into cramjam.File
    #       decompress_into appears to work fine.
    # Further todo is finding out why with pytest.raises(...) is delayed in
    # detecting the raised error. :S
    if variant_str == "blosc2" and output_type == cramjam.File:
        pytest.skip("NotImplementedError for blosc2 into File")

    variant = getattr(cramjam, variant_str)

    # Setup input
    if input_type == "numpy":
        input = np.frombuffer(raw_data, dtype=np.uint8)
    elif input_type == cramjam.File:
        path = tmp_path_factory.mktemp("tmp").joinpath("input.txt")
        path.touch()
        input = cramjam.File(str(path))
        input.write(raw_data)
        input.seek(0)
    elif input_type == cramjam.Buffer:
        input = cramjam.Buffer()
        input.write(raw_data)
        input.seek(0)
    else:
        input = input_type(raw_data)

    compressed = variant.compress(raw_data)
    compressed_len = len(compressed)

    # Setup output buffer
    output: Union[np.ndarray, cramjam.File, cramjam.Buffer, bytes]
    if output_type == "numpy":
        output = np.zeros(compressed_len, dtype=np.uint8)
    elif output_type == cramjam.File:
        path = tmp_path_factory.mktemp("tmp").joinpath("output.txt")
        path.touch()
        output = cramjam.File(str(path))
    elif output_type == cramjam.Buffer:
        output = cramjam.Buffer()
    else:
        output = output_type(b"0" * compressed_len)

    if (is_pypy or is_free_threaded) and isinstance(output, (bytes, memoryview)):
        pytest.skip(
            reason="Decompressing into immutable objects is not supported on PyPy or the free-threaded build"
        )

    n_bytes = variant.compress_into(input, output)

    if isinstance(output, (cramjam.File, cramjam.Buffer)):
        output.seek(0)
        output = output.read()
    elif isinstance(output, np.ndarray):
        output = output.tobytes()
    else:
        output = bytes(output)

    decompressed = variant.decompress(output[:n_bytes])
    assert same_same(raw_data, decompressed)


@pytest.mark.parametrize(
    "input_type", (bytes, bytearray, "numpy", cramjam.Buffer, cramjam.File, memoryview)
)
@pytest.mark.parametrize(
    "output_type", (bytes, bytearray, "numpy", cramjam.Buffer, cramjam.File, memoryview)
)
@pytest.mark.parametrize("variant_str", VARIANTS)
@given(raw_data=st.binary())
def test_variants_decompress_into(
    variant_str,
    input_type,
    output_type,
    tmp_path_factory,
    raw_data,
    is_pypy,
    is_free_threaded,
):
    if variant_str == "izlib" and output_type == "memoryview":
        pytest.skip("See issue https://github.com/milesgranger/cramjam/issues/193")

    variant = getattr(cramjam, variant_str)

    compressed = variant.compress(raw_data)

    # Setup input
    if input_type == "numpy":
        input = np.frombuffer(compressed, dtype=np.uint8)
    elif input_type == cramjam.File:
        path = tmp_path_factory.mktemp("tmp").joinpath("input.txt")
        path.touch()
        input = cramjam.File(str(path))
        input.write(compressed)
        input.seek(0)
    elif input_type == cramjam.Buffer:
        input = cramjam.Buffer()
        input.write(compressed)
        input.seek(0)
    else:
        input = input_type(compressed)

    # Setup output buffer
    if output_type == "numpy":
        output = np.zeros(len(raw_data), dtype=np.uint8)
    elif output_type == cramjam.File:
        path = tmp_path_factory.mktemp("tmp").joinpath("output.txt")
        path.touch()
        output = cramjam.File(str(path))
    elif output_type == cramjam.Buffer:
        output = cramjam.Buffer()
    else:
        output = output_type(b"0" * len(raw_data))

    if (is_pypy or is_free_threaded) and isinstance(output, (bytes, memoryview)):
        pytest.skip(
            reason="Decompressing into immutable objects is not supported on PyPy or the free-threaded build"
        )

    n_bytes = variant.decompress_into(input, output)
    assert n_bytes == len(raw_data)

    if isinstance(output, (cramjam.File, cramjam.Buffer)):
        output.seek(0)
        output = output.read()
    elif hasattr(output, "tobytes"):
        output = output.tobytes()
    else:
        output = bytes(output)
    assert same_same(output, raw_data)


@given(data=st.binary())
def test_variant_snappy_raw_into(data):
    """
    A little more special than other de/compress_into variants, as the underlying
    snappy raw api makes a hard expectation that its calculated len is used.
    """

    compressed = cramjam.snappy.compress_raw(data)
    compressed_size = cramjam.snappy.compress_raw_max_len(data)
    compressed_buffer = np.zeros(compressed_size, dtype=np.uint8)
    n_bytes = cramjam.snappy.compress_raw_into(data, compressed_buffer)
    assert n_bytes == len(compressed)

    decompressed_buffer = np.zeros(len(data), dtype=np.uint8)
    n_bytes = cramjam.snappy.decompress_raw_into(
        compressed_buffer[:n_bytes].tobytes(), decompressed_buffer
    )
    assert n_bytes == len(data)

    assert same_same(decompressed_buffer[:n_bytes], data)


@given(data=st.binary())
def test_variant_lz4_block_into(data):
    """
    A little more special than other de/compress_into variants, as the underlying
    snappy raw api makes a hard expectation that its calculated len is used.
    """

    compressed = cramjam.lz4.compress_block(data)
    compressed_size = cramjam.lz4.compress_block_bound(data)
    compressed_buffer = np.zeros(compressed_size, dtype=np.uint8)
    n_bytes = cramjam.lz4.compress_block_into(data, compressed_buffer)
    assert n_bytes == len(compressed)
    assert same_same(compressed, compressed_buffer[:n_bytes])

    decompressed_buffer = np.zeros(len(data), dtype=np.uint8)
    n_bytes = cramjam.lz4.decompress_block_into(
        compressed_buffer[:n_bytes].tobytes(), decompressed_buffer
    )
    assert n_bytes == len(data)

    assert same_same(decompressed_buffer[:n_bytes], data)


@pytest.mark.parametrize("Obj", (cramjam.File, cramjam.Buffer))
@given(data=st.binary())
def test_dunders(Obj, tmp_path_factory, data):
    path = None
    if Obj == cramjam.File:
        path = tmp_path_factory.mktemp("tmp").joinpath("tmp.txt")
        path.touch()
        obj = Obj(str(path))
    else:
        obj = Obj()

    assert len(obj) == 0
    assert bool(obj) is False
    obj.write(data)
    assert len(obj) == len(data)
    assert bool(obj) is bool(len(data))

    assert f"len={len(data)}" in str(obj)
    if isinstance(obj, cramjam.File):
        assert f"path={path}" in str(obj)


@pytest.mark.parametrize(
    "compress_kwargs",
    (
        dict(mode="default", acceleration=1, compression=1, store_size=True),
        dict(mode="fast", acceleration=2, compression=2, store_size=False),
        dict(mode="high_compression", acceleration=3, compression=3, store_size=True),
        dict(mode="default", acceleration=5, compression=4, store_size=False),
    ),
)
def test_lz4_block(compress_kwargs):
    lz4 = cramjam.lz4

    data = b"howdy neighbor"

    # What python-lz4 outputs in block mode
    expected = b"\x0e\x00\x00\x00\xe0howdy neighbor"
    assert bytes(lz4.compress_block(data)) == expected

    # and what it does without 'store_size=True'
    expected = b"\xe0howdy neighbor"
    assert bytes(lz4.compress_block(data, store_size=False)) == expected

    # Round trip the current collection of compression kwargs
    out = lz4.decompress_block(
        lz4.compress_block(data, **compress_kwargs),
        output_len=len(data) if not compress_kwargs["store_size"] else None,
    )
    assert same_same(bytes(out), data)


@given(first=st.binary(), second=st.binary())
def test_gzip_multiple_streams(first: bytes, second: bytes):
    out1 = gzip.compress(first)
    out2 = gzip.compress(second)
    assert gzip.decompress(out1 + out2) == first + second

    # works with data compressed by std gzip lib
    out = bytes(cramjam.gzip.decompress(out1 + out2))
    assert out == first + second

    # works with data compressed by cramjam
    o1 = bytes(cramjam.gzip.compress(first))
    o2 = bytes(cramjam.gzip.compress(second))
    out = bytes(cramjam.gzip.decompress(o1 + o2))
    assert same_same(out, first + second)


@pytest.mark.parametrize(
    "mod",
    (
        cramjam.brotli,
        cramjam.bzip2,
        cramjam.deflate,
        cramjam.gzip,
        cramjam.lz4,
        cramjam.snappy,
        cramjam.zstd,
    ),
)
@given(first=st.binary(), second=st.binary())
def test_streams_compressor(mod, first: bytes, second: bytes):
    compressor = mod.Compressor()

    compressor.compress(first)
    out = bytes(compressor.flush())

    compressor.compress(second)
    out += bytes(compressor.flush())

    out += bytes(compressor.finish())
    decompressed = mod.decompress(out)
    assert same_same(bytes(decompressed), first + second)

    # just empty bytes after the first .finish()
    # same behavior as brotli.Compressor()
    assert bytes(compressor.finish()) == b""

    # compress will raise an error as the stream is completed
    with pytest.raises(cramjam.CompressionError):
        compressor.compress(b"data")


@pytest.mark.parametrize("variant_str", VARIANTS)
def test_variants_stream_decompressors(variant_str):
    variant = getattr(cramjam, variant_str)
    Decompressor = getattr(variant, "Decompressor")
    decompressor = Decompressor()

    compressed = variant.compress(b"bytes")
    for _ in range(2):
        n_bytes = decompressor.decompress(bytes(compressed))
        assert n_bytes == 5
    assert bytes(decompressor.flush()) == b"bytesbytes"
    assert bytes(decompressor.flush()) == b""

    decompressor.decompress(bytes(compressed))
    assert bytes(decompressor.finish()) == b"bytes"

    # Calling .finish renders decompressor unusable after. (API consistency with other libs)
    with pytest.raises(cramjam.DecompressionError):
        decompressor.finish()


def test_buffer_cmp():
    assert cramjam.Buffer() == cramjam.Buffer()
    assert cramjam.Buffer(b"some bytes") == cramjam.Buffer(b"some bytes")
    assert cramjam.Buffer(b"some bytes") != cramjam.Buffer(b"other bytes")



from typing import Optional
from cramjam import Buffer, BufferProtocol


def decompress(data: BufferProtocol, output_len: Optional[int] = None) -> Buffer:
    """
    Brotli decompression.

    Python Example
    --------------
    >>> cramjam.brotli.decompress(compressed_bytes, output_len=Optional[int])
    """
    ...


def compress(
    data: BufferProtocol,
    level: Optional[int] = None,
    output_len: Optional[int] = None,
) -> Buffer:
    """
    Brotli compression.

    Python Example
    --------------
    >>> cramjam.brotli.compress(b'some bytes here', level=9, output_len=Optional[int])  # level defaults to 11
    """
    ...


def compress_into(
    input: BufferProtocol,
    output: BufferProtocol,
    level: Optional[int] = None,
) -> int:
    """
    Compress directly into an output buffer.
    """
    ...


def decompress_into(
    input: BufferProtocol,
    output: BufferProtocol,
) -> int:
    """
    Decompress directly into an output buffer.
    """
    ...


class Compressor:
    """
    Brotli Compressor object for streaming compression.
    """

    def __init__(self, level: Optional[int] = None) -> None:
        """
        Initialize a new `Compressor` instance.
        """
        ...

    def compress(self, input: BufferProtocol) -> int:
        """
        Compress input into the current compressor's stream.

        Returns the number of bytes written to the stream.
        """
        ...

    def flush(self) -> Buffer:
        """
        Flush and return current compressed stream.
        """
        ...

    def finish(self) -> Buffer:
        """
        Consume the current compressor state and return the compressed stream.

        NB: The compressor will not be usable after this method is called.
        """
        ...


class Decompressor:
    """
    Brotli Decompressor object for streaming decompression.
    """

    def decompress(self, input: BufferProtocol) -> Buffer: ...
    def flush(self) -> Buffer: ...
    def finish(self) -> Buffer: ...



import platform
import sysconfig

import pytest


@pytest.fixture(scope="session")
def is_pypy():
    impl = platform.python_implementation()
    return impl.lower() == "pypy"

@pytest.fixture(scope="session")
def is_free_threaded():
    return bool(sysconfig.get_config_var("Py_GIL_DISABLED"))


def pytest_configure(config):
    config.addinivalue_line("markers", "skip_pypy: skip this test on PyPy")


def pytest_runtest_setup(item):
    if "skip_pypy" in item.keywords and platform.python_implementation() == "PyPy":
        pytest.skip("skipped on PyPy")



from typing import Optional
from cramjam import Buffer, BufferProtocol


def decompress(data: BufferProtocol, output_len: Optional[int] = None) -> Buffer:
    """
    LZ4 decompression.

    Example
    -------
    ```python
    >>> # Note, output_len is currently ignored; underlying algorithm does not support reading to slice at this time
    >>> cramjam.lz4.decompress(compressed_bytes, output_len=Optional[int])
    ```
    """
    ...


def compress(
    data: BufferProtocol, level: Optional[int] = None, output_len: Optional[int] = None
) -> Buffer:
    """
    LZ4 compression.

    Example
    -------
    ```python
    >>> # Note, output_len is currently ignored; underlying algorithm does not support reading to slice at this time
    >>> cramjam.lz4.compress(b'some bytes here', output_len=Optional[int])
    ```
    """
    ...


def compress_into(
    input: BufferProtocol, output: BufferProtocol, level: Optional[int] = None
) -> int:
    """
    Compress directly into an output buffer
    """
    ...


def decompress_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """
    Decompress directly into an output buffer
    """
    ...


def decompress_block(data: BufferProtocol, output_len: Optional[int] = None) -> Buffer:
    """
    LZ4 _block_ decompression.

    `output_len` is optional, it's the upper bound length of decompressed data; if it's not provided,
    then it's assumed `store_size=True` was used during compression and length will then be taken
    from the header, otherwise it's assumed `store_size=False` was used and no prepended size exists in input

    Example
    -------
    ```python
    >>> cramjam.lz4.decompress_block(compressed_bytes, output_len=Optional[int])
    ```

    """
    ...


def compress_block(
    data: BufferProtocol,
    output_len: Optional[int] = None,
    mode: Optional[str] = None,
    acceleration: Optional[int] = None,
    compression: Optional[int] = None,
    store_size: Optional[bool] = None,
) -> Buffer:
    """
    LZ4 _block_ compression.

    Example
    -------
    ```python
    >>> cramjam.lz4.compress_block(
    ...     b'some bytes here',
    ...     output_len=Optional[int],
    ...     mode=Option[str],
    ...     acceleration=Option[int],
    ...     compression=Option[int],
    ...     store_size=Option[bool]
    ... )
    ```
    """
    ...


def decompress_block_into(
    input: BufferProtocol,
    output: BufferProtocol,
    output_len: Optional[int] = None,
) -> int:
    """
    LZ4 _block_ decompression into a pre-allocated buffer.

    Example
    -------
    ```python
    >>> cramjam.lz4.decompress_block_into(compressed_bytes, output_buffer)
    ```
    """
    ...


def compress_block_into(
    data: BufferProtocol,
    output: BufferProtocol,
    mode: Optional[str] = None,
    acceleration: Optional[int] = None,
    store_size: Optional[bool] = None,
) -> int:
    """
    LZ4 _block_ compression into pre-allocated buffer.

    Example
    -------
    ```python
    >>> cramjam.lz4.compress_block_into(
    ...     b'some bytes here',
    ...     output=output_buffer,
    ...     mode=Option[str],
    ...     acceleration=Option[int],
    ...     compression=Option[int],
    ...     store_size=Option[bool]
    ... )
    ```

    """
    ...


def compress_block_bound(src: BufferProtocol) -> int:
    """
    Determine the size of a buffer which is guaranteed to hold the result of block compression, will error if
    data is too long to be compressed by LZ4.

    Example
    -------
    ```python
    >>> cramjam.lz4.compress_block_bound(b'some bytes here')
    ```
    """
    ...


class Compressor:
    def __init__(
        self,
        level: Optional[int] = None,
        content_checksum: Optional[bool] = None,
        block_linked: Optional[bool] = None,
    ) -> None:
        """
        Initialize a new `Compressor` instance.
        """
        ...

    def compress(self, input: bytes) -> int: ...
    def flush(self) -> Buffer: ...
    def finish(self) -> Buffer: ...


class Decompressor:
    def __init__(self, *args, **kwargs) -> None: ...
    def decompress(self, data: bytes) -> Buffer: ...



from typing import Optional
from cramjam import BufferProtocol, Buffer


def decompress(data: BufferProtocol, output_len: Optional[int] = ...) -> Buffer:
    """
    Snappy decompression.

    Python Example
    --------------
    ```python
    # bytes or bytearray; bytearray is faster
    cramjam.snappy.decompress(compressed_bytes, output_len=None)
    ```
    """
    ...


def compress(data: BufferProtocol, output_len: Optional[int] = ...) -> Buffer:
    """
    Snappy compression.

    Python Example
    --------------
    ```python
    _ = cramjam.snappy.compress(b'some bytes here')
    _ = cramjam.snappy.compress(bytearray(b'this avoids double allocation in rust side, and thus faster!'))
    ```
    """
    ...


def decompress_raw(data: BufferProtocol, output_len: Optional[int] = ...) -> Buffer:
    """
    Snappy decompression, raw
    This does not use the snappy 'framed' encoding of compressed bytes.

    Python Example
    --------------
    ```python
    cramjam.snappy.decompress_raw(compressed_raw_bytes)
    ```
    """
    ...


def compress_raw(data: BufferProtocol, output_len: Optional[int] = ...) -> Buffer:
    """
    Snappy compression raw.
    This does not use the snappy 'framed' encoding of compressed bytes.

    Python Example
    --------------
    ```python
    cramjam.snappy.compress_raw(b'some bytes here')
    ```
    """
    ...


def compress_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """Compress directly into an output buffer"""
    ...


def decompress_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """Decompress directly into an output buffer"""
    ...


def compress_raw_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """Compress raw format directly into an output buffer"""
    ...


def decompress_raw_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """Decompress raw format directly into an output buffer"""
    ...


def compress_raw_max_len(data: BufferProtocol) -> int:
    """
    Get the expected max compressed length for snappy raw compression; this is the size
    of buffer that should be passed to `compress_raw_into`
    """
    ...


def decompress_raw_len(data: BufferProtocol) -> int:
    """
    Get the decompressed length for the given data. This is the size of buffer
    that should be passed to `decompress_raw_into`
    """
    ...


class Compressor:
    """Snappy Compressor object for streaming compression"""

    def __init__(self) -> None:
        """Initialize a new `Compressor` instance."""
        ...

    def compress(self, input: bytes) -> int:
        """Compress input into the current compressor's stream."""
        ...

    def flush(self) -> Buffer:
        """Flush and return current compressed stream"""
        ...

    def finish(self) -> Buffer:
        """Consume the current compressor state and return the compressed stream
        **NB** The compressor will not be usable after this method is called."""
        ...


class Decompressor:
    """Snappy streaming Decompressor (generated via make_decompressor!)."""

    ...



from typing import Optional
from cramjam import Buffer, BufferProtocol


def decompress(data: BufferProtocol, output_len: Optional[int] = None) -> Buffer:
    """
    bzip2 decompression.

    Example
    -------
    ```python
    >>> cramjam.bzip2.decompress(compressed_bytes, output_len=Optional[int])
    ```
    """
    ...


def compress(
    data: BufferProtocol, level: Optional[int] = None, output_len: Optional[int] = None
) -> Buffer:
    """
    bzip2 compression.

    Example
    -------
    ```python
    >>> cramjam.bzip2.compress(b'some bytes here', level=6, output_len=Option[int])  # level defaults to 6
    ```
    """
    ...


def compress_into(
    input: BufferProtocol, output: BufferProtocol, level: Optional[int] = None
) -> int:
    """
    Compress directly into an output buffer
    """
    ...


def decompress_into(input: BufferProtocol, output: BufferProtocol) -> int:
    """
    Decompress directly into an output buffer
    """
    ...


class Compressor:
    def __init__(self, level: Optional[int] = None) -> None: ...
    def compress(self, input: bytes) -> int: ...
    def flush(self) -> Buffer: ...
    def finish(self) -> Buffer: ...


class Decompressor:
    def decompress(self, data: bytes) -> Buffer: ...



from _typeshed import ReadableBuffer, WriteableBuffer
from typing import Any, Union
from . import lz4, snappy, deflate, gzip, zstd, xz, zlib, brotli, bzip2, experimental


__version__: str

# TODO: all supported python versions can just use types.Buffer
BufferProtocol = Any


class CompressionError(Exception):
    """
    Cramjam specific exception representing a failed compression operation.
    """

    ...


class DecompressionError(Exception):
    """
    Cramjam specific exception representing a failed decompression operation.
    """

    ...


class Buffer(Union[ReadableBuffer, WriteableBuffer]):
    def __init__(
        self, data: BufferProtocol | None = None, copy: bool | None = True
    ) -> None:
        """
        Initialize the buffer, with any initial `data`, which is required to implement the Buffer Protocol
        Optionally, don't make a copy of the data.

        Parameters
        ----------
        data: anything implementing the buffer protocol
        copy: bool (default True)
            Make a copy of the provided data.

        Returns
        -------
        Buffer
        """
        ...

    def get_view_reference(self) -> None | Any:
        """
        Get the PyObject this Buffer is referencing as its view,
        returns None if this Buffer owns its data.
        """
        ...

    def get_view_reference_count(self) -> None | int:
        """
        Get the PyObject reference count this Buffer is referencing as its view,
        returns None if this Buffer owns its data.
        """
        ...

    def len(self) -> int:
        """
        Length of the underlying buffer
        """
        ...

    def write(self, input: BufferProtocol) -> int:
        """
        Write some bytes to the buffer, where input data can be
        anything implementing the Buffer Protocol.
        """
        ...

    def read(self, n_bytes: int | None = -1) -> bytes:
        """
        Read from the buffer in its current position,
        returns bytes; optionally specify number of bytes to read.
        """
        ...

    def readinto(self, output: BufferProtocol) -> int:
        """
        Read from the buffer in its current position, into an object implementing the Buffer Protocol.
        """
        ...

    def seek(self, position: int, whence: int | None = 0) -> int:
        """
        Seek to a position within the buffer. whence follows the same values as IOBase.seek where:
        ```bash
        0: from start of the stream
        1: from current stream position
        2: from end of the stream
        ```
        """
        ...

    def seekable(self) -> bool:
        """
        Whether the buffer is seekable; here just for compatibility, it always returns True.
        """
        ...

    def tell(self) -> int:
        """
        Give the current position of the buffer.
        """
        ...

    def set_len(self, size: int) -> None:
        """
        Set the length of the buffer. If less than current length, it will truncate to the size given;
        otherwise will be null byte filled to the size.
        """
        ...

    def truncate(self) -> None:
        """
        Truncate the buffer
        """
        ...

    def __len__(self) -> int: ...
    def __contains__(self, x: BufferProtocol) -> bool: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other) -> bool: ...
    def __bool__(self) -> bool: ...


class File[BufferProtocol]:
    def __init__(
        self,
        path: str,
        read: bool | None = None,
        write: bool | None = None,
        truncate: bool | None = None,
        append: bool | None = None,
    ) -> None:
        """
        File-like object owned on Rust side

        ### Example

        ```python
        from cramjam import File
        file = File("/tmp/file.txt", read=True, write=True, truncate=True)
        file.write(b"bytes")
        file.seek(2)
        file.read()
        b'tes'
        ```
        """
        ...

    def write(self, input: BufferProtocol) -> int:
        """
        Write some bytes to the file, where input data can be anything implementing the Buffer Protocol
        """
        ...

    def read(self, n_bytes: int | None = None) -> bytes:
        """
        Read from the file in its current position, returns `bytes`; optionally specify number of
        bytes to read.
        """
        ...

    def readinto(self, output: BufferProtocol) -> int:
        """
        Read from the file in its current position, into a object implementing the Buffer Protocol.
        """
        ...

    def seek(self, position: int, whence: int | None = 0) -> int:
        """
        Seek to a position within the file. `whence` follows the same values as [IOBase.seek](https://docs.python.org/3/library/io.html#io.IOBase.seek)
        where:
        ```bash
        0: from start of the stream
        1: from current stream position
        2: from end of the stream
        ```
        """
        ...

    def seekable(self) -> bool:
        """
        Whether the file is seekable; here just for compatibility, it always returns True.
        """
        ...

    def tell(self) -> int:
        """
        Give the current position of the file.
        """
        ...

    def set_len(self, size: int) -> None:
        """
        Set the length of the file. If less than current length, it will truncate to the size given;
        otherwise will be null byte filled to the size.
        """
        ...

    def truncate(self) -> None:
        """
        Truncate the file.
        """
        ...

    def len(self) -> int:
        """
        Length of the file in bytes
        """
        ...

    def __repr__(self) -> str: ...
    def __bool__(self) -> bool: ...
    def __len__(self) -> int: ...



"""
Test decompressing files which have been compressed from
main stream third party implementations, separate from this project.
"""

import sys
import lzma
import pathlib
from collections import namedtuple

import pytest
from hypothesis import strategies as st, given, settings
from hypothesis.extra import numpy as st_np

import cramjam
from .test_variants import same_same


@pytest.fixture
def integration_dir():
    return pathlib.Path(__file__).parent.joinpath("data/integration")


@pytest.fixture
def plaintext(integration_dir):
    return integration_dir.joinpath("plaintext.txt").read_bytes()


Variant = namedtuple("Variant", ("name", "suffix"))


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Bytes comparison fails on windows"
)
@pytest.mark.parametrize(
    "variant",
    (
        Variant("gzip", "gz"),
        Variant("bzip2", "bz2"),
        Variant("zstd", "zst"),
        Variant("brotli", "br"),
        Variant("lz4", "lz4"),
        Variant("snappy", "snappy"),
        Variant("xz", "lzma"),
    ),
)
def test_variant(variant: Variant, integration_dir: pathlib.Path, plaintext: bytes):
    file = integration_dir.joinpath(f"plaintext.txt.{variant.suffix}")
    decompress = getattr(cramjam, variant.name).decompress
    assert bytes(decompress(file.read_bytes())) == plaintext


@given(data=st.binary(min_size=1, max_size=int(1e6)))
@pytest.mark.parametrize("format", (lzma.FORMAT_ALONE, lzma.FORMAT_XZ))
def test_lzma_compat(data, format):
    # Decompress from std lzma lib
    compressed = lzma.compress(data, format=format)
    uncompressed = cramjam.xz.decompress(compressed)
    assert same_same(bytes(uncompressed), data)

    # std lzma lib can decompress us
    cjformat = (
        cramjam.xz.Format.ALONE if format == lzma.FORMAT_ALONE else cramjam.xz.Format.XZ
    )
    compressed = cramjam.xz.compress(data, format=cjformat)
    uncompressed = lzma.decompress(bytes(compressed), format=format)
    assert same_same(uncompressed, data)


@given(data=st.binary(min_size=1, max_size=int(1e6)))
@pytest.mark.parametrize("set_output_len", (True, False))
def test_lz4_decompress_block_into_non_prepended_size(data, set_output_len):
    compressed = cramjam.lz4.compress_block(data, store_size=False)

    # Check both scenarios of user explicitly providing output_len or not
    output_len = len(data) if set_output_len else None

    # If we have data, and the output buffer isn't long enough
    # it ought to fail
    if data:
        with pytest.raises(cramjam.DecompressionError):
            cramjam.lz4.decompress_block_into(
                compressed, bytearray(0), output_len=output_len
            )

        # But if we explicitly provide the right output_len,
        # while the output is less, we'll know why.
        match = f"output_len set to {len(data)}, but output is less"
        with pytest.raises(cramjam.DecompressionError, match=match):
            cramjam.lz4.decompress_block_into(
                compressed, bytearray(0), output_len=len(data)
            )

    # If it's the same length as original data, it's okay.
    out = bytearray(len(data))
    n = cramjam.lz4.decompress_block_into(compressed, out, output_len=output_len)
    assert same_same(bytes(out), data)

    # pre-allocated buffer is larger, also okay
    out = bytearray(len(compressed) * 2)
    n = cramjam.lz4.decompress_block_into(compressed, out, output_len=output_len)
    assert same_same(bytes(out[:n]), data)



import gc

import cramjam
import pytest
from cramjam import Buffer


@pytest.mark.skip_pypy
@pytest.mark.parametrize("copy", (None, True, False))
def test_buffer_view(copy):
    kwargs = dict()
    if copy is not None:
        kwargs["copy"] = copy

    data = bytearray(b"bytes")
    buf = Buffer(data, **kwargs)
    buf.write(b"0")

    if copy is False:
        assert data == b"0ytes"

    # Default is to copy, None and True behavior
    else:
        assert data == b"bytes"


@pytest.mark.skip_pypy
def test_buffer_view_raises_when_writing_past_data_length_at_once():
    data = bytearray(b"bytes")
    buf = Buffer(data, copy=False)

    # Won't write pasted underlying buffer if passed data all at once
    with pytest.raises(OSError, match="Too much to write on view"):
        buf.write(b"0" * len(data) + b"0")
    assert data == b"bytes"


@pytest.mark.skip_pypy
def test_buffer_view_raises_when_writing_past_data_length_incrementally():
    data = bytearray(b"bytes")
    buf = Buffer(data, copy=False)

    # This is okay, up to length of underlying buffer
    for _ in range(len(data)):
        buf.write(b"0")

    # Whoops, one too many bytes
    with pytest.raises(OSError, match="Too much to write on view"):
        buf.write(b"0")
    assert data == b"00000"


@pytest.mark.skip_pypy
@pytest.mark.parametrize("len", range(0, 7))
def test_buffer_view_raises_when_setting_length(len):
    data = b"bytes"
    buf = Buffer(data, copy=False)

    with pytest.raises(OSError, match="Cannot set length on unowned buffer"):
        buf.set_len(len)
    assert data == b"bytes"


@pytest.mark.skip_pypy
def test_buffer_view_raises_when_truncating():
    data = b"bytes"
    buf = Buffer(data, copy=False)

    with pytest.raises(OSError, match="Cannot truncate unowned buffer"):
        buf.truncate()
    assert data == b"bytes"


@pytest.mark.skip_pypy
@pytest.mark.parametrize("whence", (0, 1, 2))
def test_buffer_view_raises_when_write_after_bad_seek(whence):
    buf = Buffer(bytearray(b"bytes"), copy=False)

    buf.seek(2, whence=0)  # Seek forward 2 from start, also okay
    buf.seek(2, whence=1)  # Seek forward 2 from current position, okay
    buf.seek(-2, whence=2)  # Seek back -2 from end, okay
    buf.seek(0)  # Set back to start

    # Seeking 10 positions from any point is not possible with len of 5
    msg = "Bad seek: cannot seek outside bounds of unowned buffer"
    with pytest.raises(OSError, match=msg):
        buf.seek(10, whence=whence)
    buf.write(b"0")


def test_buffer_view_not_supported_on_pypy(is_pypy):
    if is_pypy:
        with pytest.raises(RuntimeError, match="copy=False not supported on PyPy"):
            Buffer(b"bytes", copy=False)


@pytest.mark.skip_pypy
def test_buffer_view_cleanup():
    n_refs = 0

    def get_buffer():
        data = bytearray(b"bytes")
        buf = cramjam.Buffer(data, copy=False)

        nonlocal n_refs
        n_refs = buf.get_view_reference_count()

        return buf

    buf = get_buffer()
    gc.collect()

    ref_count = buf.get_view_reference_count()
    assert ref_count is not None
    assert 0 < ref_count < n_refs

    # Data kept alive due to internal reference
    assert buf.read() == b"bytes"


@pytest.mark.skip_pypy
def test_buffer_view_changing_underlying_buffer_size():
    # A buffer owning it's data
    data = Buffer()
    data.write(b"bytes")

    # Our reference buffer
    buf = Buffer(data, copy=False)

    # Can write 5 bytes, no problem.
    buf.write(b"12345")

    # 6th is an issue
    with pytest.raises(IOError, match="Too much to write on view"):
        buf.write(b"6")

    # buf if we extend our data, then we can
    assert len(buf) == 5
    data.write(b"s")
    assert len(buf) == 6  # Length sync'd with underlying buffer

    assert buf.tell() == 5
    buf.write(b"6")
    assert buf.tell() == 6

    # Now, shrink the underlying data
    data.set_len(2)
    assert buf.tell() == 2  # updated to end of buffer

    # Writing fails b/c the underlying is at 2 in length now
    with pytest.raises(IOError, match="Too much to write on view"):
        buf.write(b"6")

    # Seek to 1 and write one byte
    buf.seek(1)
    buf.write(b"1")
    assert buf.tell() == 2  # back at 2
    assert len(buf) == 2


@pytest.mark.skip_pypy
def test_buffer_view_cannot_read_passed():
    data = b"bytes"
    buf = cramjam.Buffer(data, copy=False)

    # Cannot read pass length of underlying buffer
    # matches read behavior of io.BytesIO
    assert buf.read(len(data) * 2) == data

    # Cannot read pass incrementally either
    b = b""
    buf.seek(0)
    for i in range(0, 10):
        b += buf.read(i)
    assert b == data



from enum import Enum
from typing import Optional
from cramjam import BufferProtocol, Buffer


class Filter(Enum):
    """Available Filter IDs"""

    Arm: Filter
    ArmThumb: Filter
    Ia64: Filter
    Lzma1: Filter
    Lzma2: Filter
    PowerPC: Filter
    Sparc: Filter
    X86: Filter


class MatchFinder(Enum):
    """MatchFinder, used with Options.mf attribute"""

    HashChain3: MatchFinder
    HashChain4: MatchFinder
    BinaryTree2: MatchFinder
    BinaryTree3: MatchFinder
    BinaryTree4: MatchFinder


class Mode(Enum):
    """MatchFinder, used with Options.mode attribute"""

    Fast: Mode
    Normal: Mode


class Options:
    def __init__(self) -> None: ...
    def set_preset(self, preset: int) -> Options: ...
    def set_dict_size(self, dict_size: int) -> Options: ...
    def set_lc(self, lc: int) -> Options: ...
    def set_lp(self, lp: int) -> Options: ...
    def set_pb(self, pb: int) -> Options: ...
    def set_mode(self, mode: Mode) -> Options: ...
    def set_nice_len(self, nice_len: int) -> Options: ...
    def set_mf(self, mf: MatchFinder) -> Options: ...
    def set_depth(self, depth: int) -> Options: ...


class Format(Enum):
    """Possible formats"""

    AUTO: Format
    XZ: Format
    ALONE: Format
    RAW: Format


class Check(Enum):
    """Possible check configurations"""

    Crc64: Check
    Crc32: Check
    Sha256: Check
    NONE: Check


class FilterChainItem:
    def __init__(self, filter: Filter, options: Optional[Options] = None) -> None: ...


class FilterChain:
    """
    FilterChain, similar to the default Python XZ filter chain which is a list of dicts
    """

    ...

    def __init__(self) -> None: ...
    def append_filter(self, filter_chain_item: FilterChainItem) -> None: ...


def compress(
    data: BufferProtocol,
    preset: Optional[int] = None,
    format: Optional[Format] = None,
    check: Optional[Check] = None,
    filters: Optional[FilterChain] = None,
    options: Optional[Options] = None,
    output_len: Optional[int] = None,
) -> Buffer:
    """
    LZMA compression.

    Example
    -------
    ```python
    >>> _ = cramjam.xz.compress(b'some bytes here')
    >>> # Defaults to XZ format, you can use the deprecated LZMA format like this:
    >>> _ = cramjam.xz.compress(b'some bytes here', format=cramjam.xz.Format.ALONE)
    ```
    """
    ...


def compress_into(
    input: BufferProtocol,
    output: BufferProtocol,
    preset: Optional[int] = None,
    format: Optional[Format] = None,
    check: Optional[Check] = None,
    filters: Optional[FilterChain] = None,
    options: Optional[Options] = None,
) -> Buffer:
    """
    LZMA compression.

    Compress directly into an output buffer
    """
    ...


def decompress_into(data: BufferProtocol, output: BufferProtocol) -> int:
    """
    Decompress directly into an output buffer
    """
    ...


def decompress(data: BufferProtocol, output_len: Optional[int] = None) -> Buffer:
    """
    LZMA decompression.

    Example
    -------

    ```python
    >>> # bytes or bytearray; bytearray is faster
    >>> cramjam.xz.decompress(compressed_bytes, output_len=Optional[None])
    ```
    """



import os
import pytest
import numpy as np
from hypothesis import strategies as st, given, settings
from hypothesis.extra import numpy as st_np


try:
    from cramjam import experimental
except ImportError:
    pytest.skip("experimental module not built", allow_module_level=True)
else:
    if hasattr(experimental, "blosc2"):
        blosc2 = experimental.blosc2  # type: ignore
    else:
        pytest.skip(
            "experimental module doesn't contain blosc2", allow_module_level=True
        )


settings.register_profile("local", max_examples=10)
settings.register_profile("CI", max_examples=5)

if os.getenv("CI"):
    settings.load_profile("CI")
else:
    settings.load_profile("local")


def variants(e):
    for attr in dir(e):
        # TODO: LastCodec, LastFilter, LastRegisteredCodec/Filter not supported
        if not attr.startswith("_") and not attr.lower().startswith("last"):
            yield getattr(e, attr)


@pytest.mark.parametrize("codec", variants(blosc2.Codec), ids=lambda v: str(v))
@pytest.mark.parametrize("filter", variants(blosc2.Filter), ids=lambda v: str(v))
@pytest.mark.parametrize("clevel", variants(blosc2.CLevel), ids=lambda v: str(v))
@given(data=st_np.arrays(st_np.scalar_dtypes(), shape=st.integers(0, 10_000)))
def test_roundtrip_chunk(data, codec, filter, clevel):
    compressed = blosc2.compress_chunk(data, clevel=clevel, filter=filter, codec=codec)
    decompressed = blosc2.decompress_chunk(compressed)
    assert data.tobytes() == bytes(decompressed)


@pytest.mark.parametrize("codec", variants(blosc2.Codec), ids=lambda v: str(v))
@pytest.mark.parametrize("filter", variants(blosc2.Filter), ids=lambda v: str(v))
@pytest.mark.parametrize("clevel", variants(blosc2.CLevel), ids=lambda v: str(v))
@given(data=st_np.arrays(st_np.scalar_dtypes(), shape=st.integers(0, 10_000)))
def test_roundtrip_chunk_into(data, codec, filter, clevel):
    kwargs = dict(clevel=clevel, filter=filter, codec=codec)
    nbytes_compressed = len(blosc2.compress_chunk(data, **kwargs))

    compressed = np.empty(
        blosc2.max_compressed_len(len(data.tobytes())), dtype=np.uint8
    )
    nbytes = blosc2.compress_chunk_into(data, compressed, **kwargs)

    decompressed = np.empty(len(data.tobytes()) * 2, dtype=np.uint8)
    nbytes = blosc2.decompress_chunk_into(compressed[:nbytes], decompressed)
    assert nbytes == len(data.tobytes())
    np.array_equal(data, np.frombuffer(decompressed[:nbytes], dtype=data.dtype))
